#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ajdefine.h"
#include "ajassert.h"
#include "ajexcept.h"
#include "ajmem.h"

extern void ajDebug(char *fmt, ...);

static const Except_T Mem_Failed = {
  "Allocation failed, insufficient memory available"
};
static const Except_T Mem_Badcount = {
  "Allocation bad byte count"
};

static ajlong memAlloc = 0;
static ajlong memFree  = 0;
static ajlong memCount = 0;
static ajlong memResize = 0;
static ajlong memResizeCount = 0;
static ajlong memTotal = 0;
static ajlong memZero = 0;

/* @func ajMemAlloc ***********************************************************
**
** Allocates memory using malloc, and fails with an error message if
** unsuccessful.
**
** @param [r] nbytes [ajlong] Number of bytes required
** @param [r] file [const char*] Source file name, generated by a macro.
** @param [r] line [ajint] Source line number, generated by a macro.
** @return [void*] Successfully allocated memory, or NULL on failure.
** @@
******************************************************************************/

void* ajMemAlloc (ajlong nbytes, const char* file, ajint line) {
  void *ptr;

  if (nbytes <= 0) {
#ifdef HAVE_JAVA
      fprintf(stderr,"Attempt to allocate <=0 bytes");
      exit(-1);
#else
      ajExceptRaise(&Mem_Badcount, file, line);
#endif
  }

  ptr = malloc(nbytes);
#ifdef HAVE_JAVA
  if(ptr == NULL)
  {
      fprintf(stderr,"Memory allocation failed in ajMemAlloc");
      exit(-1);
  }
#else
  if (ptr == NULL) {
    if (file == NULL)
      AJRAISE(Mem_Failed);
    else
      ajExceptRaise(&Mem_Failed, file, line);
  }
#endif

  memAlloc += nbytes;
  memCount++;
  memTotal++;

  return ptr;
}

/* @func ajMemCalloc **********************************************************
**
** Allocates memory using calloc for an array of elements,
** and fails with an error message if unsuccessful.
**
** @param [r] count [ajlong] Number of elements required
** @param [r] nbytes [ajlong] Number of bytes required per element
** @param [r] file [const char*] Source file name, generated by a macro.
** @param [r] line [ajint] Source line number, generated by a macro.
** @return [void*] Successfully allocated memory, or NULL on failure.
** @@
******************************************************************************/

void* ajMemCalloc (ajlong count, ajlong nbytes,
		   const char* file, ajint line) {

  void *ptr;

  assert(count > 0);
  assert(nbytes > 0);

  ptr = calloc(count, nbytes);
#ifdef HAVE_JAVA
  if(ptr == NULL)
  {
      fprintf(stderr,"Memory allocation failed in ajMemCalloc");
      exit(-1);
  }
#else
  if (ptr == NULL) {
    if (file == NULL)
      AJRAISE(Mem_Failed);
    else
      ajExceptRaise(&Mem_Failed, file, line);
  }
#endif

  memAlloc += (count*nbytes);
  memCount++;
  memTotal++;

  return ptr;
}

/* @func ajMemCalloc0 *********************************************************
**
** Allocates memory using malloc for an array of elements,
** and fails with an error message if unsuccessful.
**
** The memory is initialised to zero.
**
** @param [r] count [ajlong] Number of elements required
** @param [r] nbytes [ajlong] Number of bytes required
** @param [r] file [const char*] Source file name, generated by a macro.
** @param [r] line [ajint] Source line number, generated by a macro.
** @return [void*] Successfully allocated memory, or NULL on failure.
** @@
******************************************************************************/

void* ajMemCalloc0 (ajlong count, ajlong nbytes,
		    const char* file, ajint line) {
  void *ptr;

  ptr = ajMemCalloc(count, nbytes, file, line);
  (void) memset (ptr, 0, count*nbytes);

  memAlloc += (count*nbytes);
  memZero += (count*nbytes);
  memCount++;
  memTotal++;

  return ptr;
}

/* @func ajMemFree ************************************************************
**
** Frees memory using 'free' and zeroes the pointer. Ignores NULL
** (uninitialized) pointers.
**
** @param [u] ptr [void*] Pointer to memory previously allocated with 'malloc'
** @param [r] file [const char*] Source file name, generated by a macro.
** @param [r] line [ajint] Source line number, generated by a macro.
** @return [void]
** @@
******************************************************************************/

void ajMemFree (void* ptr, const char* file, ajint line) {
  if (ptr) {
    free(ptr);
    memCount--;
    memFree++;
    ptr = NULL;
  }
  return;
}

/* @func ajMemResize **********************************************************
**
** Resizes previously allocated memory, and ensures data is copied to
** the new location if it is moved.
**
** If the pointer is new then new memory is allocated automatically.
**
** @param [u] ptr [void*] Pointer to memory previously allocated with 'malloc'
** @param [r] nbytes [ajlong] Number of bytes required
** @param [r] file [const char*] Source file name, generated by a macro.
** @param [r] line [ajint] Source line number, generated by a macro.
** @return [void*] Successfully allocated memory, or NULL on failure.
**                 A copy of the new value of 'ptr'
** @@
******************************************************************************/

void* ajMemResize (void* ptr, ajlong nbytes,
		   const char* file, ajint line) {

  assert(nbytes > 0);

  if (ptr == NULL) {
    ptr = ajMemAlloc(nbytes, file, line);
    return ptr;
  }

  ptr = realloc(ptr, nbytes);

#ifdef HAVE_JAVA
  if(ptr == NULL)
  {
      fprintf(stderr,"Memory allocation failed in ajMemResize");
      exit(-1);
  }
#else
  if (ptr == NULL) {
    if (file == NULL)
      AJRAISE(Mem_Failed);
    else
      ajExceptRaise(&Mem_Failed, file, line);
  }
#endif

  memResize += nbytes;
  memResizeCount++;

  return ptr;
}

/* @func ajMemArrB ************************************************************
**
** Creates an AjBool array.
** Use AJFREE to free the memory when no longer needed.
**
** @param [r] size [size_t] Number of array elements.
** @return [ajint*] Newly allocated array.
** @@
******************************************************************************/

ajint* ajMemArrB (size_t size) {
  return AJCALLOC (size, sizeof(AjBool));
}

/* @func ajMemArrI ************************************************************
**
** Creates an integer array.
** Use AJFREE to free the memory when no longer needed.
**
** @param [r] size [size_t] Number of array elements.
** @return [ajint*] Newly allocated array.
** @@
******************************************************************************/

ajint* ajMemArrI (size_t size) {
  return AJCALLOC (size, sizeof(ajint));
}

/* @func ajMemArrF ************************************************************
**
** Creates a float array.
** Use AJFREE to free the memory when no longer needed.
**
** @param [r] size [size_t] Number of array elements.
** @return [float*] Newly allocated array.
** @@
******************************************************************************/

float* ajMemArrF (size_t size) {
  return AJCALLOC (size, sizeof(float));
}

/* @func ajMemStat ************************************************************
**
** Prints a summary of memory usage with debug calls
**
** @param [r] title [char*] Title for this summary
** @return [void]
** @@
******************************************************************************/

void ajMemStat (char* title) {

  static ajlong statAlloc = 0;
  static ajlong statCount = 0;
  static ajlong statFree = 0;
  static ajlong statResize = 0;
  static ajlong statResizeCount = 0;
  static ajlong statTotal = 0;
  static ajlong statZero = 0;

  ajDebug ("Memory usage since last call %s:\n", title);
  ajDebug ("Memory usage (bytes): %ld allocated, %ld reallocated %ld zeroed\n",
	   memAlloc - statAlloc, memResize - statResize, memZero - statZero);
  ajDebug ("Memory usage (number): %ld allocates, "
	   "%ld frees, %ld resizes, %ld in use\n",
	   memTotal - statTotal, memFree - statFree,
	   memResizeCount - statResizeCount, memCount - statCount);

  statAlloc = memAlloc;
  statCount = memCount;
  statFree = memFree;
  statResize = memResize;
  statResizeCount = memResizeCount;
  statTotal = memTotal;
  statZero = memZero;

  return;
}
/* @func ajMemExit ************************************************************
**
** Prints a summary of memory usage with debug calls
**
** @return [void]
** @@
******************************************************************************/

void ajMemExit (void) {

  ajDebug ("Memory usage (bytes): %ld allocated, %ld reallocated %ld zeroed\n",
	   memAlloc, memResize, memZero);
  ajDebug ("Memory usage (number): %ld allocates, "
	   "%ld frees, %ld resizes, %ld in use\n",
	   memTotal, memFree, memResizeCount, memCount);

  return;
}
